#ifndef BSP_Q7S_DEVICES_PLOCSUPVHELPER_H_
#define BSP_Q7S_DEVICES_PLOCSUPVHELPER_H_

#include <string>

#include "OBSWConfig.h"
#include "fsfw/devicehandlers/CookieIF.h"
#include "fsfw/objectmanager/SystemObject.h"
#include "fsfw/osal/linux/BinarySemaphore.h"
#include "fsfw/returnvalues/HasReturnvaluesIF.h"
#include "fsfw/tasks/ExecutableObjectIF.h"
#include "fsfw_hal/linux/uart/UartComIF.h"
#include "linux/devices/devicedefinitions/PlocSupervisorDefinitions.h"
#ifdef XIPHOS_Q7S
#include "bsp_q7s/memory/SdCardManager.h"
#endif

/**
 * @brief   Helper class for supervisor of PLOC intended to accelerate large data transfers between
 *          the supervisor and the OBC.
 * @author  J. Meier
 */
class PlocSupvHelper : public SystemObject, public ExecutableObjectIF, public HasReturnvaluesIF {
 public:
  static const uint8_t SUBSYSTEM_ID = SUBSYSTEM_ID::PLOC_SUPV_HELPER;

  //! [EXPORT] : [COMMENT] update failed
  static const Event SUPV_UPDATE_FAILED = MAKE_EVENT(0, severity::LOW);
  //! [EXPORT] : [COMMENT] update successful
  static const Event SUPV_UPDATE_SUCCESSFUL = MAKE_EVENT(1, severity::LOW);
  //! [EXPORT] : [COMMENT] Terminated update procedure by command
  static const Event TERMINATED_UPDATE_PROCEDURE = MAKE_EVENT(2, severity::LOW);
  //! [EXPORT] : [COMMENT] Requesting event buffer was successful
  static const Event SUPV_EVENT_BUFFER_REQUEST_SUCCESSFUL = MAKE_EVENT(3, severity::LOW);
  //! [EXPORT] : [COMMENT] Requesting event buffer failed
  static const Event SUPV_EVENT_BUFFER_REQUEST_FAILED = MAKE_EVENT(4, severity::LOW);
  //! [EXPORT] : [COMMENT] Terminated event buffer request by command
  //! P1: Number of packets read before process was terminated
  static const Event SUPV_EVENT_BUFFER_REQUEST_TERMINATED = MAKE_EVENT(5, severity::LOW);
  //! [EXPORT] : [COMMENT] Communication interface returned failure when trying to send the command
  //! to the supervisor
  //! P1: Return value returned by the communication interface sendMessage function
  //! P2: Internal state of supervisor helper
  static const Event SUPV_SENDING_COMMAND_FAILED = MAKE_EVENT(6, severity::LOW);
  //! [EXPORT] : [COMMENT] Request receive message of communication interface failed
  //! P1: Return value returned by the communication interface requestReceiveMessage function
  //! P2: Internal state of supervisor helper
  static const Event SUPV_HELPER_REQUESTING_REPLY_FAILED = MAKE_EVENT(7, severity::LOW);
  //! [EXPORT] : [COMMENT] Reading receive message of communication interface failed
  //! P1: Return value returned by the communication interface readingReceivedMessage function
  //! P2: Internal state of supervisor helper
  static const Event SUPV_HELPER_READING_REPLY_FAILED = MAKE_EVENT(8, severity::LOW);
  //! [EXPORT] : [COMMENT] Did not receive acknowledgement report
  //! P1: Number of bytes missing
  //! P2: Internal state of MPSoC helper
  static const Event SUPV_MISSING_ACK = MAKE_EVENT(9, severity::LOW);
  //! [EXPORT] : [COMMENT] Supervisor did not receive execution report
  //! P1: Number of bytes missing
  //! P2: Internal state of supervisor helper
  static const Event SUPV_MISSING_EXE = MAKE_EVENT(10, severity::LOW);
  //! [EXPORT] : [COMMENT] Supervisor received acknowledgment failure report
  //! P1: Internal state of supervisor helper
  static const Event SUPV_ACK_FAILURE_REPORT = MAKE_EVENT(11, severity::LOW);
  //! [EXPORT] : [COMMENT] Execution report failure
  //! P1:
  static const Event SUPV_EXE_FAILURE_REPORT = MAKE_EVENT(12, severity::LOW);
  //! [EXPORT] : [COMMENT] Supervisor expected acknowledgment report but received space packet with
  //! other apid P1: Apid of received space packet P2: Internal state of supervisor helper
  static const Event SUPV_ACK_INVALID_APID = MAKE_EVENT(13, severity::LOW);
  //! [EXPORT] : [COMMENT] Supervisor helper expected execution report but received space packet
  //! with other apid P1: Apid of received space packet P2: Internal state of supervisor helper
  static const Event SUPV_EXE_INVALID_APID = MAKE_EVENT(14, severity::LOW);
  //! [EXPORT] : [COMMENT] Failed to receive acknowledgment report
  //! P1: Return value
  //! P2: Apid of command for which the reception of the acknowledgment report failed
  static const Event ACK_RECEPTION_FAILURE = MAKE_EVENT(15, severity::LOW);
  //! [EXPORT] : [COMMENT] Failed to receive execution report
  //! P1: Return value
  //! P2: Apid of command for which the reception of the execution report failed
  static const Event EXE_RECEPTION_FAILURE = MAKE_EVENT(16, severity::LOW);

  PlocSupvHelper(object_id_t objectId);
  virtual ~PlocSupvHelper();

  ReturnValue_t initialize() override;
  ReturnValue_t performOperation(uint8_t operationCode = 0) override;

  ReturnValue_t setComIF(UartComIF* uartComfIF_);
  void setComCookie(CookieIF* comCookie_);

  /**
   * @brief   Starts update procedure
   *
   * @param file            File containing the update data
   * @param memoryId        ID of the memory where to write to
   * @param startAddress    Address where to write data
   *
   * @return  RETURN_OK if successful, otherwise error return value
   */
  ReturnValue_t startUpdate(std::string file, uint8_t memoryId, uint32_t startAddress);

  /**
   * @brief Calling this function will initiate the procedure to request the event buffer
   */
  ReturnValue_t startEventbBufferRequest(std::string path);

  /**
   * @brief   Can be used to interrupt a running data transfer.
   */
  void stopProcess();

 private:
  static const uint8_t INTERFACE_ID = CLASS_ID::PLOC_SUPV_HELPER;

  //! [EXPORT] : [COMMENT] File accidentally close
  static const ReturnValue_t FILE_CLOSED_ACCIDENTALLY = MAKE_RETURN_CODE(0xA0);
  //! [EXPORT] : [COMMENT] Process has been terminated by command
  static const ReturnValue_t PROCESS_TERMINATED = MAKE_RETURN_CODE(0xA1);
  //! [EXPORT] : [COMMENT] Received command with invalid pathname
  static const ReturnValue_t PATH_NOT_EXISTS = MAKE_RETURN_CODE(0xA2);
  //! [EXPORT] : [COMMENT] Expected event buffer TM but received space packet with other APID
  static const ReturnValue_t EVENT_BUFFER_REPLY_INVALID_APID = MAKE_RETURN_CODE(0xA3);

  static const uint16_t CRC16_INIT = 0xFFFF;
  // Event buffer reply will carry 24 space packets with 1016 bytes and one space packet with
  // 192 bytes
  static const uint8_t NUM_EVENT_BUFFER_PACKETS = 25;
  static const size_t SIZE_EVENT_BUFFER_FULL_PACKET = 1024;
  static const size_t SIZE_EVENT_BUFFER_LAST_PACKET = 200;
  static const uint32_t CRC_EXECUTION_TIMEOUT = 60000;

  struct Update {
    uint8_t memoryId;
    uint32_t startAddress;
    // Absolute name of file containing update data
    std::string file;
    // Size of update
    uint32_t length;
    uint32_t crc;
  };

  struct Update update;

  struct EventBufferRequest {
    std::string path = "";
    // Default name of file where event buffer data will be written to. Timestamp will be added to
    // name when new file is created
    std::string filename = "event-buffer.bin";
  };

  EventBufferRequest eventBufferReq;

  enum class InternalState { IDLE, UPDATE, REQUEST_EVENT_BUFFER };

  InternalState internalState = InternalState::IDLE;

  BinarySemaphore semaphore;
#ifdef XIPHOS_Q7S
  SdCardManager* sdcMan = nullptr;
#endif
  uint8_t commandBuffer[supv::MAX_COMMAND_SIZE];

  bool terminate = false;

  /**
   * Communication interface responsible for data transactions between OBC and Supervisor.
   */
  UartComIF* uartComIF = nullptr;
  // Communication cookie. Must be set by the supervisor Handler
  CookieIF* comCookie = nullptr;

  bool timestamping = true;

  // Remembers APID to know at which command a procedure failed
  uint16_t rememberApid = 0;

  ReturnValue_t performUpdate();
  ReturnValue_t performEventBufferRequest();
  ReturnValue_t handlePacketTransmission(SpacePacket& packet,
                                         uint32_t timeoutExecutionReport = 1000);
  ReturnValue_t sendCommand(SpacePacket& packet);
  /**
   * @brief Function which reads form the communication interface
   *
   * @param data            Pointer to buffer where read data will be written to
   * @param raedBytes       Actual number of bytes read
   * @param requestBytes    Number of bytes to read
   */
  ReturnValue_t receive(uint8_t* data, size_t* readBytes, size_t requestBytes);
  ReturnValue_t handleAck();
  ReturnValue_t handleExe(uint32_t timeout = 1000);
  /**
   * @brief Handles reading of TM packets from the communication interface
   *
   * @param tmPacket        Pointer to space packet where received data will be written to
   * @param reaminingBytes  Number of bytes to read in the space packet
   * @param timeout         Receive timeout in milliseconds
   */
  ReturnValue_t handleTmReception(supv::TmPacket* tmPacket, size_t remainingBytes,
                                  uint32_t timeout = 1000);
  ReturnValue_t prepareUpdate();
  ReturnValue_t eraseMemory();
  // Calculates CRC over image. Will be used for verification after update writing has
  // finished.
  ReturnValue_t calcImageCrc();
  ReturnValue_t handleCheckMemoryCommand();
  /**
   * @brief Return size of file with name filename
   *
   * @param filename
   *
   * @return The size of the file
   */
  uint32_t getFileSize(std::string filename);
  ReturnValue_t handleEventBufferReception();
};

#endif /* BSP_Q7S_DEVICES_PLOCSUPVHELPER_H_ */
